#include <stdlib.h>
#include <glut.h>
#include <math.h>
#include "HeightMap.h"
#include "Object.h"
#include "TGA.H"

HeightMap::HeightMap() {
	callstackNumber = 0;
	terrainHeights = NULL;
	terrainColors = NULL;
	terrainNormals = NULL;
	terrainGridWidth = 0;
	terrainGridLength = 0; 
}

HeightMap::~HeightMap() {
	HeightMapClear();
}

void
HeightMap::loadMap() {
	// for now, programmatically generate until near finish
	float xOffset = 0;
	float yOffset = 0;
	float zOffset = 0;

	GLuint terrainDL;
	float startW,startL;
	int i,j,aux;

	// compute the initial point of the terrain on the XZ plane
	startW = terrainGridWidth / 2.0 - terrainGridWidth;
	startL = - terrainGridLength / 2.0 + terrainGridLength;
	
	// Create the id for the display list
	terrainDL = glGenLists(1);

	// create the display list
	glNewList(terrainDL,GL_COMPILE);
	
	// test for normals
	if (terrainNormals != NULL) {
		glColorMaterial(GL_FRONT, GL_DIFFUSE);
		glEnable(GL_COLOR_MATERIAL);
	}

	// generate n-1 strips, where n = terrainGridLength
	// for each vertex test if colors and normals are enabled
	for (i = 0 ; i < terrainGridLength-1; i++) {
		glBegin(GL_TRIANGLE_STRIP);
		for (j = 0;j < terrainGridWidth; j++) {
			aux = 3*((i+1)*terrainGridWidth + j);
			if (terrainColors != NULL) 
				glColor3f(terrainColors[aux],
					terrainColors[aux+1],
					terrainColors[aux+2]);
			if (terrainNormals != NULL)
				glNormal3f(terrainNormals[aux],
					terrainNormals[aux+1],
					terrainNormals[aux+2]);
			glVertex3f(
				startW + j + xOffset,
				terrainHeights[(i+1)*terrainGridWidth + (j)] + yOffset,
				startL - (i+1) + zOffset);

			aux = 3*(i*terrainGridWidth + j);
			if (terrainColors != NULL) 
				glColor3f(terrainColors[aux],
					terrainColors[aux+1],
					terrainColors[aux+2]);
			if (terrainNormals != NULL)
				glNormal3f(terrainNormals[aux],
					terrainNormals[aux+1],
					terrainNormals[aux+2]);
			glVertex3f(
				startW + j + xOffset, 
				terrainHeights[(i)*terrainGridWidth + j] + yOffset,
				startL - i + zOffset);					
		}
		glEnd();
	}
	glEndList();

	// return the list index so that the application can use it
	callstackNumber = terrainDL;
}

void HeightMap::HeightMapClear() {
	terrainGridWidth = 0;
	terrainGridLength = 0; 
	delete terrainHeights;
	delete terrainColors;
	delete terrainNormals;
	terrainHeights = NULL;
	terrainColors = NULL;
	terrainNormals = NULL;
}

    
int 
HeightMap::terrainLoadFromImage(char *filename, int normals) {

	tgaInfo *info;
	int mode,aux;
	float pointHeight;

	// if a terrain already exists, destroy it.
	if (terrainHeights != NULL)
		HeightMapClear();
		
	// load the image, using the tgalib
	info = tgaLoad(filename);
	
	// check to see if the image was properly loaded
	// remember: only greyscale, RGB or RGBA noncompressed images
	if (info->status != TGA_OK)
		return(info->status);

	// if the image is RGB, convert it to greyscale.
	// mode will store the image's number of components
	mode = info->pixelDepth / 8;
	if (mode == 3) {
		tgaRGBtogreyscale(info);
		mode = 1;
	}
	
	// set the width and height of the terrain
	terrainGridWidth = info->width;
	terrainGridLength = info->height;

	// alocate memory for the terrain, and check for errors
	terrainHeights = (float *)malloc(terrainGridWidth * 
					terrainGridLength * 
					sizeof(float));
	if (terrainHeights == NULL)
		return(TERRAIN_ERROR_MEMORY_PROBLEM);

	// allocate memory for the normals, and check for errors
	if (normals) {
		terrainNormals = (float *)malloc(terrainGridWidth * 
						terrainGridLength * 
						sizeof(float) * 3);
		if (terrainNormals == NULL)
			return(TERRAIN_ERROR_MEMORY_PROBLEM);
	}
	else
			terrainNormals = NULL;

	// if mode = RGBA then allocate memory for colors, and check for errors
	if (mode == 4) {
		terrainColors = (float *)malloc(terrainGridWidth * 
						terrainGridLength * 
						sizeof(float)*3);
		if (terrainColors == NULL)
			return(TERRAIN_ERROR_MEMORY_PROBLEM);
	}
	else
		terrainColors = NULL;

	// fill arrays
	for (int i = 0 ; i < terrainGridLength; i++)
		for (int j = 0;j < terrainGridWidth; j++) {
		
			// compute the height as a value between 0.0 and 1.0
			aux = mode*(i*terrainGridWidth + j);
			pointHeight = info->imageData[aux+(mode-1)] / 256.0;
			terrainHeights[i*terrainGridWidth + j] = pointHeight;
			
			// if mode = RGBA then fill the colors array as well
			if (mode==4) {
				terrainColors[3*(i*terrainGridWidth + j)] =
				 	info->imageData[aux] / 256.0;
				terrainColors[3*(i*terrainGridWidth + j)+1] =
				 	info->imageData[aux+1]/256.0;
				terrainColors[3*(i*terrainGridWidth + j)+2] = 
					info->imageData[aux+2]/256.0;
			}
		}
	// if we want normals then compute them		
	if (normals) terrainComputeNormals();
		
	// free the image's memory 
	tgaDestroy(info);
	
	return(TERRAIN_OK); 
}

int 
HeightMap::getCallStackNumber() {
	return callstackNumber;
}

int 
HeightMap::terrainScale(float min,float max) {

	float amp,aux,min1,max1,height;
	int total,i;

	if (terrainHeights == NULL)
		return(TERRAIN_ERROR_NOT_INITIALISED);

	if (min > max) {
		aux = min;
		min = max;
		max = aux;
	}

	amp = max - min;
	total = terrainGridWidth * terrainGridLength;

	min1 = terrainHeights[0];
	max1 = terrainHeights[0];
	for(i=1;i < total ; i++) {
		if (terrainHeights[i] > max1)
			max1 = terrainHeights[i];
		if (terrainHeights[i] < min1)
			min1 = terrainHeights[i];
	}
	for(i=0;i < total; i++) {
		height = (terrainHeights[i] - min1) / (max1-min1);
		terrainHeights[i] = height * amp - min;
	}
printf("min,max,min1,max1,amp,height %f %f %f %f %f %f\n",min,max,min1,max1,amp,height);
	if (terrainNormals != NULL)	terrainComputeNormals();
	return(TERRAIN_OK);
}

float 
HeightMap::terrainGetHeight(int x, int z) {

	int xt,zt;

	if (terrainHeights == NULL) 
			return(0.0);

	xt = x + terrainGridWidth /2;
	zt = terrainGridWidth - (z + terrainGridLength /2);

	if ((xt > terrainGridWidth) || (zt > terrainGridLength) || (xt < 0) || (zt < 0))
		return(0.0);

	return(terrainHeights[zt * terrainGridWidth + xt]);
}

float*
HeightMap::terrainCrossProduct(int x1,int z1,int x2,int z2,int x3,int z3) {

	float *auxNormal,v1[3],v2[3];
		
	v1[0] = x2-x1; 
	v1[1] = -terrainHeights[z1 * terrainGridWidth + x1] 
			+ terrainHeights[z2 * terrainGridWidth + x2];
	v1[2] = z2-z1; 


	v2[0] = x3-x1; 
	v2[1] = -terrainHeights[z1 * terrainGridWidth + x1] 
			+ terrainHeights[z3 * terrainGridWidth + x3];
	v2[2] = z3-z1; 

	auxNormal = (float *)malloc(sizeof(float)*3);

	auxNormal[2] = v1[0] * v2[1] - v1[1] * v2[0];
	auxNormal[0] = v1[1] * v2[2] - v1[2] * v2[1];
	auxNormal[1] = v1[2] * v2[0] - v1[0] * v2[2];

	return(auxNormal);
}

void 
HeightMap::terrainNormalize(float *v) {

	double d,x,y,z;

	x = v[0];
	y = v[1];
	z = v[2];
	
	d = sqrt((x*x) + (y*y) + (z*z));

	v[0] = x / d;
	v[1] = y / d;
	v[2] = z / d;
}

void 
HeightMap::terrainAddVector(float *a, float *b) {

	a[0] += b[0];
	a[1] += b[1];
	a[2] += b[2];
}

void 
HeightMap::terrainComputeNormals() {

	float *norm1,*norm2,*norm3,*norm4; 
	int i,j,k;
	
	if (terrainNormals == NULL)
		return;

	for(i = 0; i < terrainGridLength; i++)
		for(j = 0; j < terrainGridWidth; j++) {
			norm1 = NULL;
			norm2 = NULL;
			norm3 = NULL;
			norm4 = NULL;

			// normals for the four corners
			if (i == 0 && j == 0) {
				norm1 = terrainCrossProduct(0,0, 0,1, 1,0);	
				terrainNormalize(norm1);				
			}
			else if (j == terrainGridWidth-1 && i == terrainGridLength-1) {
				norm1 = terrainCrossProduct(i,j, j,i-1, j-1,i);	
				terrainNormalize(norm1);				
			}
			else if (j == 0 && i == terrainGridLength-1) {
				norm1 = terrainCrossProduct(i,j, j,i-1, j+1,i);	
				terrainNormalize(norm1);				
			}
			else if (j == terrainGridWidth-1 && i == 0) {
				norm1 = terrainCrossProduct(i,j, j,i+1, j-1,i);	
				terrainNormalize(norm1);				
			}

			// normals for the borders
			else if (i == 0) {
				norm1 = terrainCrossProduct(j,0, j-1,0, j,1);
				terrainNormalize(norm1);
				norm2 = terrainCrossProduct(j,0,j,1,j+1,0);
				terrainNormalize(norm2);
			}
			else if (j == 0) {
				norm1 = terrainCrossProduct(0,i, 1,i, 0,i-1);
				terrainNormalize(norm1);
				norm2 = terrainCrossProduct(0,i, 0,i+1, 1,i);
				terrainNormalize(norm2);
			}
			else if (i == terrainGridLength) {
				norm1 = terrainCrossProduct(j,i, j,i-1, j+1,i);
				terrainNormalize(norm1);
				norm2 = terrainCrossProduct(j,i, j+1,i, j,i-1);
				terrainNormalize(norm2);
			}
			else if (j == terrainGridWidth) {
				norm1 = terrainCrossProduct(j,i, j,i-1, j-1,i);
				terrainNormalize(norm1);
				norm2 = terrainCrossProduct(j,i, j-1,i, j,i+1);
				terrainNormalize(norm2);
			}

			// normals for the interior
			else {
				norm1 = terrainCrossProduct(j,i, j-1,i, j,i+1);
				terrainNormalize(norm1);
				norm2 = terrainCrossProduct(j,i, j,i+1, j+1,i);
				terrainNormalize(norm2);
				norm3 = terrainCrossProduct(j,i, j+1,i, j,i-1);
				terrainNormalize(norm3);
				norm4 = terrainCrossProduct(j,i, j,i-1, j-1,i);
				terrainNormalize(norm4);
			}
			if (norm2 != NULL) {
				terrainAddVector(norm1,norm2);
				free(norm2);
			}
			if (norm3 != NULL) {
				terrainAddVector(norm1,norm3);
				free(norm3);
			}
			if (norm4 != NULL) {
				terrainAddVector(norm1,norm4);
				free(norm4);
			}
			terrainNormalize(norm1);
			norm1[2] = - norm1[2];
			for (k = 0; k< 3; k++) 
				terrainNormals[3*(i*terrainGridWidth + j) + k] = norm1[k];
			free(norm1);
		}

}